import unittest
from PcnnLib.network_lib.network import Network
from PcnnLib.network_lib.neuron_factories.full_neuron_factory import FullNeuronFactory


class TestNetworkLinking(unittest.TestCase):

    def assert_connections(self, target_neuron, neuron):
        target_input_connections = target_neuron.input_connections
        target_output_connections = target_neuron.output_connections
        input_connections = neuron.input_connections
        output_connections = neuron.output_connections
        common_connections = set(target_input_connections) & set(output_connections)
        self.assertEqual(len(common_connections), 1)
        common_connections = set(target_output_connections) & set(input_connections)
        self.assertEqual(len(common_connections), 1)

    def test_neurons_connected_correctly(self):
        nn = Network(True)
        neuron_factory = FullNeuronFactory(1, 1, 1, 1, 1, 1, 1)
        nn.init_layer((10, 10), neuron_factory)
        width = len(nn._neurons)
        height = len(nn._neurons)
        self.assert_connections(nn._neurons[0][0], nn._neurons[1][1])
        self.assert_connections(nn._neurons[0][height - 1], nn._neurons[1][height - 1])
        self.assert_connections(nn._neurons[width - 1][height - 1], nn._neurons[width - 2][height - 2])



from itertools import chain


class TimeSignature:
    def __init__(self):
        self._signature = []
        self._thresholds = []

    def add_iteration(self, stat_record):
        sgn_value = sum(chain.from_iterable(stat_record.spikes))
        self._signature.append(sgn_value)
        self._thresholds.append(stat_record.total_threshold)

    def get_signature(self):
        return self._signature

    def get_thresholds(self):
        return self._thresholds



from PcnnLib.image_preprocessing.image_converters import to_binary_spikes, to_grayscale, read_img, spikes_to_img, save_img, to_spikes, \
    to_normal_spikes
from PcnnLib.network_lib.network import Network
import argparse
import pyclustering


def main():
    parser = argparse.ArgumentParser(description='Filter image background.')
    parser.add_argument('--input', dest='input_path', help='input file path')
    parser.add_argument('--output', dest='output_path', help='output file path')

    args = parser.parse_args()
    img = read_img(args.input_path)
    img = to_grayscale(img)
    img_list = to_binary_spikes(img)
    nn = Network(1, 1, 30, 1, 0.1, 0.1, 0.8, True)
    nn.init_layer([len(img_list), len(img_list[0])])
    for i in range(4):
        print(f'iteration: {i}')
        output_img_list = nn.iterate(img_list)
        #output_img_list = denormalize_spikes(output_img_list)
        output_img = spikes_to_img(output_img_list)
        save_img(output_img, args.output_path + "_iter_" + str(i) + ".jpg")


if __name__ == "__main__":
    main()




import numpy
from PIL import Image


def read_img(path):
    return Image.open(path)


def save_img(img, path):
    return img.save(path)


def resize_img(img, size):
    return img.resize((size[1], size[0]), Image.ANTIALIAS)


def spikes_to_img(spikes):
    """Convert 2d matrix of spikes to image.
    # Arguments
        spikes: 2d list with values between 0 and 255
    # Returns
        PIL image
    """
    arr = numpy.asarray([numpy.array(r, dtype=numpy.uint8) for r in spikes])
    img = Image.fromarray(arr)
    return img


def denormalize_spikes_scaled(spikes):
    arr = numpy.asarray([numpy.array(r) for r in spikes])
    normalize_item_vec = numpy.vectorize(lambda i: i * 255)
    return normalize_item_vec(arr)


def to_grayscale(img):
    return img.convert('L')


def to_spikes(img):
    return numpy.asarray(img)


def to_normal_spikes(img):
    spikes = to_spikes(img)
    maximum_stimulus = spikes.max()
    minimum_stimulus = spikes.min()
    delta = maximum_stimulus - minimum_stimulus
    normalize_item_vec = numpy.vectorize(lambda i: (float(i) - minimum_stimulus) / delta)
    spikes = normalize_item_vec(spikes)
    return spikes


def to_binary_spikes(img, fixed_threshold=None):
    spikes = to_spikes(img)
    if fixed_threshold is not None:
        normalize_item_vec = numpy.vectorize(lambda i: 1 if i > fixed_threshold else 0)
        spikes = normalize_item_vec(spikes)
        return spikes
    maximum_stimulus = spikes.max()
    minimum_stimulus = spikes.min()
    delta = maximum_stimulus - minimum_stimulus
    average = delta / 2 + minimum_stimulus
    normalize_item_vec = numpy.vectorize(lambda i: 1 if i > average else 0)
    spikes = normalize_item_vec(spikes)
    return spikes


def apply_mask(img, mask):
    """Apply a mask on an image. The area outside the mask turns black.
    # Arguments
        image: PIL image
        mask: PIL mask
    # Returns
        PIL masked image
    """
    img_arr = numpy.asarray(img)
    result = []
    for i in range(len(mask) - 1):
        row = []
        for j in range(len(mask[0]) - 1):
            row.append(numpy.array([0, 0, 0], dtype=numpy.uint8) if mask[i][j] == 0 else img_arr[i][j])
        result.append(row)
    return spikes_to_img(result)



import os


def get_filename(path):
    base = os.path.basename(path)
    return os.path.splitext(base)[0]


def find_file(dir_path, filename):
    for file in os.listdir(dir_path):
        if get_filename(file) == filename:
            return os.path.join(dir_path, file)
    return None



from PcnnLib.network_lib.neurons.simple_neuron import SimpleNeuron


class SimpleNeuronFactory:
    def __init__(self,
                 linking_scale,
                 linking_strength,
                 threshold_scale,
                 threshold_decay_step):
        self.LINKING_SCALE = linking_scale
        self.LINKING_STRENGTH = linking_strength
        self.THRESHOLD_SCALE = threshold_scale
        self.THRESHOLD_DECAY_STEP = threshold_decay_step

    def create_neurons(self, shape):
        return [
            [
                SimpleNeuron(
                    self.LINKING_SCALE,
                    self.LINKING_STRENGTH,
                    self.THRESHOLD_SCALE,
                    self.THRESHOLD_DECAY_STEP)
                for j in range(shape[1])
            ] for i in range(shape[0])
        ]



from PcnnLib.network_lib.neurons.full_neuron import FullNeuron


class FullNeuronFactory:
    def __init__(self,
                 feeding_scale,
                 linking_scale,
                 threshold_scale,
                 linking_strength,
                 af,
                 al,
                 at):
        self.FEEDING_SCALE = feeding_scale
        self.LINKING_SCALE = linking_scale
        self.THRESHOLD_SCALE = threshold_scale
        self.LINKING_STRENGTH = linking_strength
        self.AF = af
        self.AL = al
        self.AT = at

    def create_neurons(self, shape):
        return [
            [
                FullNeuron(
                    self.FEEDING_SCALE,
                    self.LINKING_SCALE,
                    self.THRESHOLD_SCALE,
                    self.LINKING_STRENGTH,
                    self.AF,
                    self.AL,
                    self.AT)
                for j in range(shape[1])
            ] for i in range(shape[0])
        ]



import unittest
import numpy as np

from PcnnLib.tools.spike_collector import collect_spikes


class TestTools(unittest.TestCase):

    def test_spike_collector_collect_correctly(self):
        spike_packages = [
            [
                [0, 1, 1],
                [1, 1, 1],
                [1, 1, 0]
            ],
            [
                [1, 1, 0],
                [1, 1, 1],
                [0, 1, 0]
            ],
            [
                [0, 1, 0],
                [1, 1, 1],
                [0, 1, 1]
            ]
        ]
        actual_collected_spikes = collect_spikes(spike_packages, 2)
        expected_collected_spikes = [
            [0, 1, 0],
            [1, 1, 1],
            [0, 1, 0]
        ]
        np.testing.assert_array_equal(actual_collected_spikes, expected_collected_spikes)


from PcnnLib.network_lib.connection import Connection
import math

from PcnnLib.network_lib.statistic_record import StatisticRecord


class Network:
    def __init__(self, fast_linking, metrics):
        self.FAST_LINKING = fast_linking
        self._metrics = metrics
        self._neurons = None

    def init_layer(self, layer_shape, neuron_factory):
        self._neurons = neuron_factory.create_neurons(layer_shape)
        frames = []
        for i in range(layer_shape[0]):
            for j in range(layer_shape[1]):
                if j - 1 >= 0:
                    neuron00 = self._neurons[i - 1][j - 1] if (i - 1 >= 0) else None
                    neuron10 = self._neurons[i][j - 1]
                    neuron20 = self._neurons[i + 1][j - 1] if (i + 1 < layer_shape[0]) else None
                else:
                    neuron00 = None
                    neuron10 = None
                    neuron20 = None

                neuron01 = self._neurons[i - 1][j] if (i - 1 >= 0) else None
                neuron11 = self._neurons[i][j]
                neuron21 = self._neurons[i + 1][j] if (i + 1 < layer_shape[0]) else None

                if j + 1 < layer_shape[1]:
                    neuron02 = self._neurons[i - 1][j + 1] if (i - 1 >= 0) else None
                    neuron12 = self._neurons[i][j + 1]
                    neuron22 = self._neurons[i + 1][j + 1] if (i + 1 < layer_shape[0]) else None
                else:
                    neuron02 = None
                    neuron12 = None
                    neuron22 = None

                frame = [
                    [neuron00, neuron01, neuron02],
                    [neuron10, neuron11, neuron12],
                    [neuron20, neuron21, neuron22]
                ]
                frames.append(frame)

        for frame in frames:
            weight = 1 / math.sqrt(2)
            self.__connect(frame[1][1], frame[0][1], 1)
            self.__connect(frame[1][1], frame[1][0], 1)
            self.__connect(frame[1][1], frame[1][2], 1)
            self.__connect(frame[1][1], frame[2][1], 1)

            self.__connect(frame[1][1], frame[0][0], weight)
            self.__connect(frame[1][1], frame[2][0], weight)
            self.__connect(frame[1][1], frame[0][2], weight)
            self.__connect(frame[1][1], frame[2][2], weight)

    def iterate(self, spikes, mask_spikes):
        result = []
        for i in range(len(spikes)):
            row = []
            for j in range(len(spikes[0])):
                spike = self._neurons[i][j].process_spike(spikes[i][j])
                row.append(spike)
            result.append(row)

        if self.FAST_LINKING:
            changes = True
            while changes:
                changes = False
                next_result = []
                for i in range(len(spikes)):
                    row = []
                    for j in range(len(spikes[0])):
                        curr_res = self._neurons[i][j].process_spike_fast_linking()
                        row.append(curr_res)
                        changes |= result[i][j] != curr_res
                    next_result.append(row)
                result = next_result

        total_threshold = 0
        for i in range(len(spikes)):
            for j in range(len(spikes[0])):
                total_threshold += self._neurons[i][j].recompute_threshold(result[i][j])

        for metric in self._metrics:
            metric.add_iteration(StatisticRecord(result, mask_spikes, total_threshold))

        return result

    def __connect(self, main_neuron, neuron, weight):
        if neuron is not None:
            conn = Connection(weight, 1, 0)
            main_neuron.connect_output(conn)
            neuron.connect_input(conn)



import os

from PcnnLib.helpers.file_helpers import find_file
from PcnnLib.image_preprocessing.image_converters import read_img, resize_img, to_grayscale, to_normal_spikes


class ImagesPipeline:
    def __init__(self, image_size, flow_images_dir_path, flow_image_names, flow_masks_dir_path=None):
        self.IMAGE_SIZE = image_size
        self._flow_image_names = flow_image_names
        self._image_index = 0
        self._flow_images_dir_path = flow_images_dir_path
        self._flow_masks_dir_path = flow_masks_dir_path

    def __iter__(self):
        return self

    def __next__(self):
        if self._image_index > len(self._flow_image_names) - 1:
            return None
        filename = self._flow_image_names[self._image_index]
        self._image_index += 1
        img_path = os.path.join(self._flow_images_dir_path, filename)
        img = read_img(img_path).convert("RGB")
        img = resize_img(img, self.IMAGE_SIZE)
        img = to_grayscale(img)
        spikes = to_normal_spikes(img)

        if self._flow_masks_dir_path is None:
            mask_spikes = None
        else:
            mask_path = find_file(self._flow_masks_dir_path, filename)
            if mask_path is None:
                return spikes, None
            img = read_img(mask_path).convert("RGB")
            img = resize_img(img, self.IMAGE_SIZE)
            mask_spikes = to_normal_spikes(img)

        return spikes, mask_spikes



from PcnnLib.image_preprocessing.directory_pipeline import DirectoryPipeline
from PcnnLib.image_preprocessing.images_pipeline import ImagesPipeline
from PcnnLib.metrics.mse import Mse
from PcnnLib.network_lib.network import Network
from PcnnLib.network_lib.neuron_factories.full_neuron_factory import FullNeuronFactory
import numpy as np


class GridSearch:
    def __init__(self):
        self._values = []

    def find_best_value(self, img_size, img_dir, mask_dir, neuron_params, target_param_name, min_value, max_value, step,
                        image_names=None):
        best_mse = None
        best_i = 0
        for i in np.arange(min_value, max_value, step):
            neuron_params[target_param_name] = i
            neuron_factory = FullNeuronFactory(**neuron_params)
            mse_metric = Mse()
            network = Network(fast_linking=True, metrics=[mse_metric])
            network.init_layer(img_size, neuron_factory)
            img_pipeline = DirectoryPipeline(img_size, img_dir, mask_dir) if image_names is None else ImagesPipeline(
                img_size, img_dir, image_names, mask_dir)

            for spikes, mask_spikes in img_pipeline:
                network.iterate(spikes, mask_spikes)

            current_mse = mse_metric.get_mean_metric()
            self._values.append(current_mse)
            if best_mse is None or current_mse > best_mse:
                best_mse = current_mse
                best_i = i
        return best_i, best_mse

    def get_values(self):
        return self._values



import numpy as np


def collect_spikes(spikes_packages, threshold):
    i_n = len(spikes_packages[0])
    j_n = len(spikes_packages[0][0])
    frequences = np.zeros((i_n, j_n))

    for spikes in spikes_packages:
        for i in range(i_n):
            for j in range(j_n):
                frequences[i][j] += spikes[i][j]

    result = np.zeros((i_n, j_n))
    for i in range(i_n):
        for j in range(j_n):
            if frequences[i][j] > threshold:
                result[i][j] = 1

    return result



from PcnnLib.helpers.file_helpers import get_filename, find_file
from PcnnLib.image_preprocessing.image_converters import read_img, resize_img, to_grayscale, to_normal_spikes


class DirectoryPipeline:
    def __init__(self, image_size, flow_images_dir_path, flow_masks_dir_path=None):
        self.IMAGE_SIZE = image_size
        import os
        self._dir_iterator = os.scandir(flow_images_dir_path)
        self._flow_masks_dir_path = flow_masks_dir_path

    def __iter__(self):
        return self

    def __next__(self):
        entry = next(self._dir_iterator)
        while entry.is_file() is False:
            entry = next(self._dir_iterator)

        img = read_img(entry.path).convert("RGB")
        img = resize_img(img, self.IMAGE_SIZE)
        img = to_grayscale(img)
        spikes = to_normal_spikes(img)

        if self._flow_masks_dir_path is None:
            mask_spikes = None
        else:
            mask_path = find_file(self._flow_masks_dir_path, get_filename(entry.path))
            if mask_path is None:
                return spikes, None
            img = read_img(mask_path).convert("RGB")
            img = resize_img(img, self.IMAGE_SIZE)
            mask_spikes = to_normal_spikes(img)

        return spikes, mask_spikes



import unittest

from PcnnLib.network_lib.connection import Connection
from PcnnLib.network_lib.neurons.full_neuron import FullNeuron


class TestNeuron(unittest.TestCase):

    def test_neuron_all_stimulus_spike(self):
        neuron = FullNeuron(0.125, 0.125, 12, 1, 0.1, 0.1, 0.1)
        for _ in range(8):
            connection = Connection(1, 1, 1)
            neuron.connect_input(connection)
        stimulus = 1
        result = neuron.process_spike(stimulus)
        self.assertEqual(result, 1)
        neuron.recompute_threshold(result)
        result = neuron.process_spike(stimulus)
        self.assertEqual(result, 0)

    def test_neuron_only_feeding_stimulus_spike(self):
        neuron = FullNeuron(1, 1, 1.9, 1, 0.1, 0.1, 0.1)
        for _ in range(8):
            connection = Connection(1, 1, 0)
            neuron.connect_input(connection)
        stimulus = 1
        result = neuron.process_spike(stimulus)
        self.assertEqual(result, 1)
        neuron.recompute_threshold(result)
        result = neuron.process_spike(stimulus)
        self.assertEqual(result, 1)



from PcnnLib.network_lib.connection import Connection


def decompose_connections(connections):
    return [(conn.spike, conn.weight_l, conn.weight_f) for conn in connections]


class SimpleNeuron:
    def __init__(self, linking_scale, linking_strength, threshold_scale, threshold_decay_step):
        self.internal_state = 0
        self.threshold = 0
        self._self_connection = Connection(1, 1, 0)
        self._input_connections = [self._self_connection]
        self._output_connections = [self._self_connection]
        self.feeding_result = 0
        self.linking_result = 0
        self.LINKING_SCALE = linking_scale
        self.THRESHOLD_SCALE = threshold_scale
        self.LINKING_STRENGTH = linking_strength
        self.threshold_decay_step = threshold_decay_step

    def connect_input(self, input_connection):
        self._input_connections.append(input_connection)

    def connect_output(self, output_connection):
        self._output_connections.append(output_connection)

    def process_spike(self, input_stimulus):
        self.feeding_result = input_stimulus
        self.linking_result = self.LINKING_SCALE * sum(connection.spike * connection.weight_l for connection in self._input_connections)
        self.internal_state = self.feeding_result * (1 + self.LINKING_STRENGTH * self.linking_result)
        spike = 1 if (self.internal_state > self.threshold) else 0
        for conn in self._output_connections:
            conn.spike = spike
        return spike

    def process_spike_fast_linking(self):
        self.linking_result = self.LINKING_SCALE * sum(connection.spike * connection.weight_l for connection in self._input_connections)
        self.internal_state = self.feeding_result * (1 + self.LINKING_STRENGTH * self.linking_result)
        spike = 1 if (self.internal_state > self.threshold) else 0
        for conn in self._output_connections:
            conn.spike = spike
        return spike

    def recompute_threshold(self, spike):
        self.threshold -= self.threshold_decay_step
        if self.threshold < 0:
            self.threshold = 0

        if spike > 0:
            if self.threshold < 1:
                self.threshold = 1
            self.threshold *= self.THRESHOLD_SCALE
        return self.threshold



class StatisticRecord:
    def __init__(self, spikes, mask_spikes, total_threshold):
        self.spikes = spikes
        self.mask_spikes = mask_spikes
        self.total_threshold = total_threshold



import numpy as np


class Mse:
    def __init__(self):
        self._results = []

    def add_iteration(self, stat_record):
        if stat_record.mask_spikes is None:
            return
        mse_sum = 0
        total_number = 0
        for i in range(len(stat_record.spikes)):
            for j in range(len(stat_record.spikes[0])):
                mse_sum += (stat_record.spikes[i][j] - stat_record.mask_spikes[i][j])**2
                total_number += 1

        self._results.append(1 / total_number * mse_sum)

    def get_metrics(self):
        return self._results

    def get_mean_metric(self):
        return np.mean(self._results)



from setuptools import setup
from setuptools import find_packages
import requests
import os


RELEASES_URL = "https://gitlab.com/api/v4/projects/16643049/releases"


if os.environ.get('CI_COMMIT_TAG'):
    version = os.environ['CI_COMMIT_TAG']
else:
    version = requests.get(RELEASES_URL).json()[0]["name"]

setup(
    name='PcnnLib',
    packages=find_packages(),
    version=version,
    license='MIT',
    description='Pulse-coupled neural network library',
    author='Anton Zotov',
    author_email='anton@zotov.online',
    url='https://gitlab.com/trikster/pcnnlib',
    download_url='https://gitlab.com/trikster/pcnnlib',
    keywords=['pcnn', 'pulse', 'neural', 'impulse'],
    install_requires=[
        'pillow',
        'numpy'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.8'
    ],
)



import os
import unittest

from PcnnLib.image_preprocessing.image_converters import read_img, to_binary_spikes, apply_mask, to_grayscale, to_spikes


class TestImageConverters(unittest.TestCase):
    def test_neurons_connected_correctly(self):
        img = read_img("./data/image.jpg")
        mask = to_grayscale(read_img("./data/mask.jpg"))
        expected_masked_img = read_img("./data/masked_img.jpg")
        mask_spikes = to_binary_spikes(mask)

        apply_mask(img, mask_spikes).save("./data/temp.jpg")
        actual_masked_img = read_img("./data/temp.jpg")

        self.assertEqual(actual_masked_img.tobytes(), expected_masked_img.tobytes())

    def tearDown(self):
        os.remove("./data/temp.jpg")



import math

from PcnnLib.network_lib.connection import Connection


def decompose_connections(connections):
    return [(conn.spike, conn.weight_l, conn.weight_f) for conn in connections]


class FullNeuron:
    def __init__(self, feeding_scale, linking_scale, threshold_scale, linking_strength, af, al, at):
        self.internal_state = 0
        self.threshold = 0
        self._self_connection = Connection(1, 1, 0)
        self._input_connections = [self._self_connection]
        self._output_connections = [self._self_connection]
        self.feeding_result = 0
        self.linking_result = 0
        self.LINKING_SCALE = linking_scale
        self.FEEDING_SCALE = feeding_scale
        self.THRESHOLD_SCALE = threshold_scale
        self.LINKING_STRENGTH = linking_strength
        self.AF = af
        self.AL = al
        self.AT = at

    def connect_input(self, input_connection):
        self._input_connections.append(input_connection)

    def connect_output(self, output_connection):
        self._output_connections.append(output_connection)

    def process_spike(self, input_stimulus):
        dec_connections = decompose_connections(self._input_connections)
        self.feeding_result = math.exp(-self.AF) * self.feeding_result + input_stimulus + self.FEEDING_SCALE * sum(
            r * wf for r, wl, wf in dec_connections)
        self.linking_result = math.exp(-self.AL) * self.linking_result + self.LINKING_SCALE * sum(
            r * wl for r, wl, wf in dec_connections)
        self.internal_state = self.feeding_result * (1 + self.LINKING_STRENGTH * self.linking_result)
        spike = 1 if (self.internal_state > self.threshold) else 0
        for conn in self._output_connections:
            conn.spike = spike
        return spike

    def process_spike_fast_linking(self):
        dec_connections = decompose_connections(self._input_connections)
        self.linking_result = self.LINKING_SCALE * sum(r * wl for r, wl, wf in dec_connections)
        self.internal_state = self.feeding_result * (1 + self.LINKING_STRENGTH * self.linking_result)
        spike = 1 if (self.internal_state > self.threshold) else 0
        for conn in self._output_connections:
            conn.spike = spike
        return spike

    def recompute_threshold(self, spike):
        self.threshold = math.exp(-self.AT) * self.threshold + self.THRESHOLD_SCALE * spike
        return self.threshold
